/* usogres -- a replacement to postmaster tool distribute multiple postgres
 * Copyright (C) 2000-2002  HOSOKAWA Tetsuichi
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: Socket.cc,v 1.8 2001/04/18 07:14:43 hosokawa Exp $
 */

#ifdef WIN32
#include <winsock.h>
#else
#include <unistd.h>
#endif

#ifdef HAVE_ERROR_H
#  include <error.h>
#endif
#include <errno.h>

#include "Socket.h"

#ifdef WIN32
#define O_ASYNC(func,arg) \
    if (_beginthread((FuncPtr)func,0,arg) < 0) {\
        message(LOG_ERR,"_beginthread error err=%d\n",errno);\
        func(arg);\
    }
#endif

#if !defined(O_ASYNC)
#  include <sys/file.h>
#  if defined(O_ASYNC)
#    undef O_ASYNC
#  endif
#  define O_ASYNC  FASYNC
#  if defined(O_NDELAY)
#    undef O_NDELAY
#  endif
#  define O_NDELAY  FNDELAY
#endif

SockAddr_in::SockAddr_in(int port, const char* sock_addr, int family)
{
	memset(this, '\0', sizeof(SockAddr_in));
	sin_family = family;
	sin_addr.s_addr = inet_addr(sock_addr);
	sin_port = htons(port);
}

int Socket::read(char* buf, int nbytes)
{
#ifdef WIN32
	return recv(sockfd, buf, nbytes,0);
#else
	return ::read(sockfd, buf, nbytes);
#endif
}

int Socket::write(const char* buf, int nbytes)
{
#ifdef WIN32
	return send(sockfd, buf, nbytes,0);
#else
	return ::write(sockfd, buf, nbytes);
#endif
}

int Socket::readn(char* buf, int nbytes)
{
	int nleft, nread;

	nleft = nbytes;
	while (nleft > 0) {
#ifdef WIN32
		nread = recv(sockfd, buf, nbytes,0);
#else
		nread = read(buf, nleft);
#endif
		if (nread < 0) {
			return nread;
		} else if (nread == 0) {
			break;
		}
		nleft -= nread;
		buf += nread;
	}
	return nbytes - nleft;
}

int Socket::writen(const char* buf, int nbytes)
{
	int nleft, nwritten;

	nleft = nbytes;
	while (nleft > 0) {
#ifdef WIN32
		nwritten = send(sockfd,buf, nleft,0);
#else
		nwritten = write(buf, nleft);
#endif
		if (nwritten <= 0) {
			return nwritten;
		}
		nleft -= nwritten;
		buf += nwritten;
	}
	return nbytes - nleft;
}

Socket::Socket(int port, const char* sock_addr, int type, int family, int protocol)
  : sa(port, sock_addr, family)
{
	sockfd = ::socket(family, type, protocol);
}

Socket::~Socket()
{
	close();
}

int Socket::set_option(int level, int optname)
{
	int on = 1;
#ifdef WIN32
	return ::setsockopt(sockfd,level,optname,(const char *)&on,sizeof(on));
#else
	return ::setsockopt(sockfd, level, optname, (void *)&on, sizeof(on));
#endif
}

int Socket::set_non_blocking()
{
	int flags;
	int myflag = 0;

#ifdef WIN32
	return ioctlsocket(sockfd, FIONBIO,(unsigned long *)&myflag);
#else
	flags = fcntl(sockfd, F_GETFL);

#ifdef O_NONBLOCK
	myflag = O_NONBLOCK; /* POSIX version */
#elif defined(O_NDELAY)
	myflag = O_NDELAY;   /* old non-POSIX version */
#endif
	flags |= myflag;
	return fcntl(sockfd, F_SETFL, flags);
#endif	/* WIN32 */
}

int Socket::set_async()
{
#ifdef WIN32
	int myflag = 0;
	return ioctlsocket(sockfd, FIOASYNC,(unsigned long *)&myflag);
#else
	return ::fcntl(sockfd, F_SETFL, O_ASYNC);
#endif
}

int Socket::set_send_buffer_size(int size)
{
#ifdef WIN32
	return ::setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, (const char *)&size, sizeof(size));
#else
	return ::setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, (void *)&size, sizeof(size));
#endif
}

int Socket::set_recv_buffer_size(int size)
{
#ifdef WIN32
	return ::setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, (const char *)&size, sizeof(size));
#else
	return ::setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, (void *)&size, sizeof(size));
#endif
}

int Socket::bind()
{
	return ::bind(sockfd, (const sockaddr*)(&sa), sizeof(sa));
}

int Socket::listen(int backlog)
{
	return ::listen(sockfd, backlog);
}

#ifdef WIN32
int Socket::Accept()
{
	int sa_size = sizeof(sa);
	return accept(sockfd, (sockaddr*)(&sa), &sa_size);
}
#else
int Socket::accept()
{
	socklen_t sa_size = sizeof(sa);
	return ::accept(sockfd, (sockaddr*)(&sa), &sa_size);
}
#endif

int Socket::connect()
{
	return ::connect(sockfd, (const sockaddr*)(&sa), sizeof(sa));
}

#ifdef WIN32
int _stdcall closesocket (SOCKET s);
#endif

int Socket::close()
{
	int ret = -1;
	if (sockfd >= 0) {
#ifdef WIN32
		ret = closesocket(sockfd);
#else
		ret = ::close(sockfd);
#endif
	}
	sockfd = -1;
	return ret;
}

int Socket::is_valid()
{
	return sockfd != -1;
}
