/* usogres -- a replacement to postmaster tool distribute multiple postgres
 * Copyright (C) 2000-2002  HOSOKAWA Tetsuichi
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: protocol.cc,v 1.3 2001/05/14 06:25:02 hosokawa Exp $
 */

#include <sys/types.h>
#ifdef WIN32
#include "../config.h"
#else
#include <netinet/in.h>
#endif
#include <stdio.h>
#include "protocol.h"
#include "exception.h"
#include "debug.h"

// initial_size must be a multiple of sizeof(int)
const int initial_size = sizeof(int) * 16;
const int multiple_factor = sizeof(int);

void dump_protocol(char* buf, int size)
{
	PgsqlProtocol* ppp = 0;
	for (int i = 0; i < size; i++) {
		if (ppp == 0) {
			ppp = new PgsqlProtocol();
		}
		if (ppp->add_byte(buf[i])) {
			ppp->dump();
			delete ppp;
			ppp = 0;
		}
	}
	if (ppp != 0) {
		delete ppp;
	}
}

PgsqlProtocol::PgsqlProtocol(Msg protocol_type)
  : length(0),
	buf_length(0),
	buf(0),
	type(protocol_type)
{
	buf = new char[initial_size];
	if (buf == 0) {
		throw MemoryException(__FILE__, __LINE__);
	}
	buf_length = initial_size;
}

PgsqlProtocol::~PgsqlProtocol()
{
	delete [] buf;
}

bool PgsqlProtocol::add_byte(unsigned char ch)
{
	if (length == buf_length) {
		buf_length *= multiple_factor;
		char* p_new = new char[buf_length];
		if (p_new == 0) {
			throw MemoryException(__FILE__, __LINE__);
		}
		int* p_src = reinterpret_cast<int*>(buf);
		int* p_dst = reinterpret_cast<int*>(p_new);
		for (unsigned int i = 0; i < length / sizeof(int); i++) {
			*p_dst++ = *p_src++;
		}
		delete [] buf;
		buf = p_new;
	}

	buf[length++] = ch;
	return is_completed();
}

PgsqlProtocol::Msg PgsqlProtocol::get_protocol_type()
{
	if (type != Unknown) {
		return type;
	}

	static Msg msg_table[] = {
		NotificationResponse,	// 'A'
		BinaryRow,				// 'B'
		CompletedResponse,		// 'C'
		AsciiRow,				// 'D'
		ErrorResponse,			// 'E'
		FunctionCall,			// 'F'
		CopyInResponse,			// 'G'
		CopyOutResponse,		// 'H'
		EmptyQueryResponse,		// 'I'
		Unknown,				// 'J'
		BackendKeyData,			// 'K'
		Unknown,				// 'L'
		Unknown,				// 'M'
		NoticeResponse,			// 'N'
		Unknown,				// 'O'
		CursorResponse,			// 'P'
		Query,					// 'Q'
		Unknown,				// 'R'
		Unknown,				// 'S'
		RowDescription,			// 'T'
		Unknown,				// 'U'
		Unknown,				// 'V'
		Unknown,				// 'W'
		Terminate,				// 'X'
		Unknown,				// 'Y'
		ReadyForQuery,			// 'Z'
	};
	Msg ret = Unknown;
	if (buf[0] >= 'A' && buf[0] <= 'Z') {
		ret = msg_table[buf[0] - 'A'];
	}
	if (ret == Unknown) {
		switch (buf[0]) {
		case 0:
			if (static_cast<unsigned int>(length) >= sizeof(int)) {
				int id = get_int(buf, sizeof(int));
				switch (id) {
				case 16:
					ret = CancelRequest;
					break;
				case 296:
					ret = StartupPacket;
					break;
				}
			}
			break;
		case 'J':
		case 'L':
		case 'M':
		case 'O':
		case 'S':
		case 'U':
		case 'W':
		case 'Y':
			throw InvalidProtocolException(__FILE__, __LINE__);
			break;
		case 'R':
			if (static_cast<unsigned int>(length) >= (1 + sizeof(int))) {
				int id = get_int(buf + 1, sizeof(int));
				switch (id) {
				case 0:
					ret = AuthenticationOk;
					break;
				case 1:
					ret = AuthenticationKerberosV4;
					break;
				case 2:
					ret = AuthenticationKerberosV5;
					break;
				case 3:
					ret = AuthenticationUnencryptedPassword;
					break;
				case 4:
					ret = AuthenticationEncryptedPassword;
					break;
				}
			}
			break;
		case 'V':
			if (length >= 2) {
				if (buf[1] == 'G') {
					ret = FunctionResultResponse;
				} else if (buf[1] == '0') {
					ret = FunctionVoidResponse;
				}
			}
			break;
		}
	}
	return type = ret;
}

bool PgsqlProtocol::check_fixed_length_type(int len)
{
	return length == len;
}

bool PgsqlProtocol::check_id_only()
{
	return length == 1;
}

bool PgsqlProtocol::check_id_and_string()
{
	if (length == 0) {
		throw LogicalException(__FILE__, __LINE__);
	}
	return buf[length - 1] == '\0';
}

int get_field_count(unsigned int idx)
{
	unsigned char count_table[] = { 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff };
	for (int i = 0; i < 8; i++) {
		if (count_table[i] == idx) {
			return i + 1;
		}
	}
	throw LogicalException(__FILE__, __LINE__);
	return 0;
}

bool PgsqlProtocol::check_AsciiRow()
{
	bool ret = false;

	try {
		char* p;
		for (p = buf + 1; p < buf + length && *p != '\0'; p++)
			;
		if (p < buf + length && *p == '\0') {
			int field_cnt = ((p - buf) - 2) * 8 + get_field_count(static_cast<unsigned char>(*(p - 1)));
			for (int i = 0; i < field_cnt; i++) {
				int len = get_int(p, sizeof(int));
				p += len;
			}
			if (buf + length == p) {
				ret = true;
			}
		}
	} catch (ImcompletedValueException&) {
	}

	return ret;
}

bool PgsqlProtocol::check_BinaryRow()
{
	return false;
}

bool PgsqlProtocol::check_CopyDataRows()
{
	return false;
}

bool PgsqlProtocol::check_EncryptedPasswordPacket()
{
	return false;
}

bool PgsqlProtocol::check_FunctionCall()
{
	return false;
}

bool PgsqlProtocol::check_ResultResponse()
{
	return false;
}

bool PgsqlProtocol::check_NotificationResponse()
{
	return false;
}

bool PgsqlProtocol::check_RowDescription()
{
	bool ret = false;

	if (length >= 3) {
		int cnt = get_int(buf + 1, sizeof(short));
		int idx = 0;
		char* ptr = buf + 3;
		try {
			while (idx++ < cnt) {
				ug_string str = get_string(ptr);
				ptr += str.length() + 1;
				int oid = get_int(ptr, sizeof(int));
				ptr += sizeof(oid);
				short type_size = get_int(ptr, sizeof(short));
				ptr += sizeof(type_size);
				int type_modifier = get_int(ptr, sizeof(int));
				ptr += sizeof(type_modifier);
			}
			ret = true;
		} catch (ImcompletedValueException&) {
		}
	}

	return ret;
}

void PgsqlProtocol::dump_fixed_length_type(int len)
{
	dump_buf(buf, len);
}

void PgsqlProtocol::dump_id_only()
{
	// printf(" Protocol ID:%c\n", buf[0]);
}

void PgsqlProtocol::dump_id_and_string()
{
	printf(" %s\n", buf + 1);
}

void PgsqlProtocol::dump_AsciiRow()
{
	char* p;
	for (p = buf + 1; p < buf + length && *p != '\0'; p++)
		;
	if (p < buf+ length && *p == '\0') {
		int field_cnt = ((p - buf) - 2) * 8 + get_field_count(static_cast<unsigned char>(*(p - 1)));
		printf(" FieldCount:%d\n", field_cnt);
		for (int i = 0; i < field_cnt; i++) {
			printf("  Field:%d\n", i + 1);
			int len = get_int(p, sizeof(int)) - sizeof(int);
			printf("   FieldLength:%d\n", len);
#ifdef WIN32
			std::string str(p + sizeof(int), len);
#else
			string str(p + sizeof(int), len);
#endif
			printf("   Value:%s\n", str.c_str());
			p += len + sizeof(int);
		}
	}
}

void PgsqlProtocol::dump_BinaryRow()
{
	dump_buf(buf, length);
}

void PgsqlProtocol::dump_CopyDataRows()
{
	dump_buf(buf, length);
}

void PgsqlProtocol::dump_EncryptedPasswordPacket()
{
	dump_buf(buf, length);
}

void PgsqlProtocol::dump_FunctionCall()
{
	dump_buf(buf, length);
}

void PgsqlProtocol::dump_ResultResponse()
{
	dump_buf(buf, length);
}

void PgsqlProtocol::dump_NotificationResponse()
{
	dump_buf(buf, length);
}

void PgsqlProtocol::dump_RowDescription()
{
	if (!is_completed()) {
		throw LogicalException(__FILE__, __LINE__);
	}

	int cnt = get_int(buf + 1, sizeof(short));
	printf(" Row:%d\n", cnt);
	int idx = 0;
	char* ptr = buf + 3;
	try {
		while (idx++ < cnt) {
			printf("  Row:%d\n", idx);
			ug_string str = get_string(ptr);
			ptr += str.length() + 1;
			printf("   Field:%s\n", str.c_str());
			int oid = get_int(ptr, sizeof(int));
			ptr += sizeof(int);
			printf("   Oid:%d\n", oid);
			int type_size = get_int(ptr, sizeof(short));
			ptr += sizeof(short);
			printf("   Type Size:%d\n", type_size);
			int type_modifier = get_int(ptr, sizeof(int));
			ptr += sizeof(int);
			printf("   Type Modifier:%d\n", type_modifier);
		}
	} catch (ImcompletedValueException&) {
	}
}

bool PgsqlProtocol::is_completed()
{
	bool ret = false;
	switch (get_protocol_type()) {
	case AsciiRow:
		ret = check_AsciiRow();
		break;
	case BinaryRow:
		ret = check_BinaryRow();
		break;
	case CopyDataRows:
		ret = check_CopyDataRows();
		break;
	case EncryptedPasswordPacket:
		ret = check_EncryptedPasswordPacket();
		break;
	case FunctionCall:
		ret = check_FunctionCall();
		break;
	case FunctionResultResponse:
		ret = check_ResultResponse();
		break;
	case NotificationResponse:
		ret = check_NotificationResponse();
		break;
	case RowDescription:
		ret = check_RowDescription();
		break;
	case AuthenticationOk:
	case AuthenticationKerberosV4:
	case AuthenticationKerberosV5:
	case AuthenticationUnencryptedPassword:
		ret = check_fixed_length_type(5);
		break;
	case AuthenticationEncryptedPassword:
		ret = check_fixed_length_type(7);
		break;
	case BackendKeyData:
		ret = check_fixed_length_type(9);
		break;
	case CancelRequest:
		ret = check_fixed_length_type(16);
		break;
	case FunctionVoidResponse:
		ret = check_fixed_length_type(2);
		break;
	case StartupPacket:
		ret = check_fixed_length_type(296);
		break;
	case CompletedResponse:
	case CursorResponse:
	case EmptyQueryResponse:
	case ErrorResponse:
	case NoticeResponse:
	case Query:
	case UnencryptedPasswordPacket:
		ret = check_id_and_string();
		break;
	case CopyInResponse:
	case CopyOutResponse:
	case ReadyForQuery:
	case Terminate:
		ret = check_id_only();
		break;
	default:
		break;
	}

	return ret;
}

int PgsqlProtocol::get_int(char* p, int size) const
{
	int ret;
	if (buf + length < p + size) {
		throw ImcompletedValueException(__FILE__, __LINE__);
	}
	if (size == sizeof(int)) {
		unsigned int* pi = reinterpret_cast<unsigned int*>(p);
		ret = ntohl(*pi);
	} else if (size == sizeof(short)) {
		unsigned short* pi = reinterpret_cast<unsigned short*>(p);
		ret = ntohs(*pi);
	} else {
		throw LogicalException(__FILE__, __LINE__);
	}
	return ret;
}

char PgsqlProtocol::get_byte(char* p) const
{
	if (p >= buf + length) {
		throw ImcompletedValueException(__FILE__, __LINE__);
	}
	return *p;
}

const char* PgsqlProtocol::get_string(char* ptr) const
{
	char* p;
	for (p = ptr; *p != '\0' && p < buf + length; p++) {
	}
	if (p == buf + length) {
		throw ImcompletedValueException(__FILE__, __LINE__);
	}
	return ptr;
}

void PgsqlProtocol::dump()
{
	static struct {
		Msg type;
		const char* name;
	} name_table[] = {
		{ AsciiRow, "AsciiRow" },
		{ AuthenticationOk, "AuthenticationOk" },
		{ AuthenticationKerberosV4, "AuthenticationKerberosV4" },
		{ AuthenticationKerberosV5, "AuthenticationKerberosV5" },
		{ AuthenticationUnencryptedPassword, "AuthenticationUnencryptedPassword" },
		{ AuthenticationEncryptedPassword, "AuthenticationEncryptedPassword" },
		{ BackendKeyData, "BackendKeyData" },
		{ BinaryRow, "BinaryRow" },
		{ CancelRequest, "CancelRequest" },
		{ CompletedResponse, "CompletedResponse" },
		{ CopyDataRows, "CopyDataRows" },
		{ CopyInResponse, "CopyInResponse" },
		{ CopyOutResponse, "CopyOutResponse" },
		{ CursorResponse, "CursorResponse" },
		{ EmptyQueryResponse, "EmptyQueryResponse" },
		{ EncryptedPasswordPacket, "EncryptedPasswordPacket" },
		{ ErrorResponse, "ErrorResponse" },
		{ FunctionCall, "FunctionCall" },
		{ FunctionResultResponse, "FunctionResultResponse" },
		{ FunctionVoidResponse, "FunctionVoidResponse" },
		{ NoticeResponse, "NoticeResponse" },
		{ NotificationResponse, "NotificationResponse" },
		{ Query, "Query" },
		{ ReadyForQuery, "ReadyForQuery" },
		{ RowDescription, "RowDescription" },
		{ StartupPacket, "StartupPacket" },
		{ Terminate, "Terminate" },
		{ UnencryptedPasswordPacket, "UnencryptedPasswordPacket" },
		{ Unknown, "Unknown" }
	};

	int idx;
	for (idx = 0; name_table[idx].type != Unknown; idx++) {
		if (name_table[idx].type == type) {
			break;
		}
	}
	printf("%s:\n", name_table[idx].name);
	if (is_completed()) {
		switch (get_protocol_type()) {
		case AsciiRow:
			dump_AsciiRow();
			break;
		case BinaryRow:
			dump_BinaryRow();
			break;
		case CopyDataRows:
			dump_CopyDataRows();
			break;
		case EncryptedPasswordPacket:
			dump_EncryptedPasswordPacket();
			break;
		case FunctionCall:
			dump_FunctionCall();
			break;
		case FunctionResultResponse:
			dump_ResultResponse();
			break;
		case NotificationResponse:
			dump_NotificationResponse();
			break;
		case RowDescription:
			dump_RowDescription();
			break;
		case BackendKeyData:
			printf(" PID:%d\n SKEY:%08x\n", get_int(buf + 1, 4), get_int(buf + 5, 4));
			break;
		case Query:
			printf(" Query:%s\n", buf + 1);
			break;
		case CancelRequest:
		case FunctionVoidResponse:
		case StartupPacket:
			dump_fixed_length_type(length);
			break;
		case CopyInResponse:
		case CopyOutResponse:
		case ReadyForQuery:
		case Terminate:
			dump_id_only();
			break;
		case CompletedResponse:
		case CursorResponse:
		case EmptyQueryResponse:
		case ErrorResponse:
		case NoticeResponse:
		case UnencryptedPasswordPacket:
			dump_id_and_string();
			break;
		case AuthenticationOk:
			break;
		default:
			dump_buf(buf, length);
			break;
		}
	} else {
		dump_buf(buf, length);
	}
}
