/* usogres -- a replacement to postmaster tool distribute multiple postgres
 * Copyright (C) 2000-2002  HOSOKAWA Tetsuichi
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: rc.cc,v 1.6 2001/08/30 11:16:14 hosokawa Exp $
 */

#ifdef WIN32
#include "../config.h"
#include <windows.h>
#include <iostream.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#include "rc.h"
#include "exception.h"

#ifdef WIN32
const char* personal_rc = "usogresrc";
#else
const char* personal_rc = ".usogresrc";
#endif
const int buf_size = 1024;

static bool is_valid_buf(const char* buf)
{
	bool ret = false;
	bool in_dq = false;
	if (buf[0] != '#') {
		for (unsigned int i = 0; i < strlen(buf) && ret == false; i++) {
			switch (buf[i]) {
			case ' ':
			case '\t':
				if (in_dq) {
					ret = true;
				}
				break;
			case '\"':
				in_dq = !in_dq;
				break;
			case '\r':
			case '\n':
				break;
			default:
				ret = true;
				break;
			}
		}
	}
	return ret;
}

void
UsogresRC::set_defaults()
{
	p_host = "127.0.0.1";
	p_hostorg = "127.0.0.1";
	p_hostcopy = "127.0.0.1";
	i_port = 5432;
	i_portorg = 5434;
	i_portcopy = 5436;
#ifdef WIN32
	p_logdir = ".";
	i_loglevel = 0;
#else
	p_logdir = "/var/log/usogres";
#endif
	i_max_process = 32;
}

UsogresRC::UsogresRC()
{
	FILE* fp = 0;
	char* fnbuf = new char[strlen(SYSCONFDIR) + strlen("usogresrc") + 1];
	if (fnbuf == 0) {
		fp = fopen("/etc/usogresrc", "r");
	} else {
#ifdef WIN32
		sprintf(fnbuf, "%s\\usogresrc", SYSCONFDIR);
#else
		sprintf(fnbuf, "%s/usogresrc", SYSCONFDIR);
#endif
		fp = fopen(fnbuf, "r");
#ifndef WIN32
		delete [] fnbuf;
#endif
	}

	if (fp == NULL) {
		if (errno != ENOENT) {
			set_defaults();
			return;
		}
		char* p_home = getenv("HOME");
		if (p_home == NULL) {
			set_defaults();
			return;
		}
		char* p_fnamebuf = new char[strlen(p_home) + 1 + strlen(personal_rc) + 1];
#ifdef WIN32
		sprintf(p_fnamebuf, "%s\\%s", p_home, personal_rc);
#else
		sprintf(p_fnamebuf, "%s/%s", p_home, personal_rc);
#endif
		fp = fopen(p_fnamebuf, "r");
		delete [] p_fnamebuf;
		if (fp == NULL) {
			set_defaults();
			return;
		}
	}

	char* p_buf = new char[buf_size];
	string key = "";
	while (fgets(p_buf, buf_size, fp)) {
		if (strlen(p_buf) >= buf_size - 1) {
			set_defaults();
			delete p_buf;
			return;
		}
		if (!is_valid_buf(p_buf)) {
			continue;
		}
		Token tok(p_buf);
		for (int idx = 0; idx < 3; idx++) {
			string str = tok.get_next_token();
			if (str == "") {
				break;
			}
			if (idx == 0) {
				if (str == "host" || str == "hostorg" || str == "hostcopy"
						|| str == "port" || str == "portorg" || str == "portcopy"
						|| str == "logdir"
#ifdef WIN32
						|| str == "loglevel"
#endif
						|| str == "max_process") {
					key = str;
				} else {
fprintf(stderr, "%s\n", str.c_str());
					set_defaults();
					delete p_buf;
					return;
				}
			} else if (idx == 1) {
				if (str != "=") {
					set_defaults();
					delete p_buf;
					return;
				}
			} else if (idx == 2) {
				if (key == "host") {
					p_host = str;
				} else if (key == "hostorg") {
					p_hostorg = str;
				} else if (key == "hostcopy") {
					p_hostcopy = str;
				} else if (key == "port") {
					i_port = atoi(str.c_str());
				} else if (key == "portorg") {
					i_portorg = atoi(str.c_str());
				} else if (key == "portcopy") {
					i_portcopy = atoi(str.c_str());
				} else if (key == "logdir") {
					p_logdir = str;
#ifdef WIN32
				} else if (key == "loglevel") {
					i_loglevel = atoi(str.c_str());
#endif
				} else if (key == "max_process") {
					i_max_process = atoi(str.c_str());
				}
			}
		}
	}
	delete p_buf;
}

const string UsogresRC::get_host()
{
	return p_host;
}

int UsogresRC::get_port()
{
	return i_port;
}

const string UsogresRC::get_hostorg()
{
	return p_hostorg;
}

int UsogresRC::get_portorg()
{
	return i_portorg;
}

const string UsogresRC::get_hostcopy()
{
	return p_hostcopy;
}

int UsogresRC::get_portcopy()
{
	return i_portcopy;
}

int UsogresRC::get_max_process()
{
	return i_max_process;
}

#ifdef WIN32
int UsogresRC::get_loglevel()
{
	return i_loglevel;
}
#endif

const string UsogresRC::get_logdir()
{
	return p_logdir;
}

Token::Token(const char* p)
  : str(p),
	pointer(0)
{
}

void Token::skip_white_space()
{
	for (unsigned int i = pointer; i < str.length(); i++) {
		if (str[i] != ' ' && str[i] != '\t') {
			pointer = i;
			break;
		}
	}
}

string Token::get_next_token()
{
	string ret = "";
	skip_white_space();

	bool in_dq = false;
	unsigned int i;
	for (i = pointer; i < str.length(); i++) {
		switch (str[i]) {
		case ' ':
		case '\t':
			if (in_dq) {
				ret += str[i];
				pointer = i;
			} else {
				pointer = i;
				return ret;
			}
			break;
		case '=':
			if (in_dq) {
				ret += str[i];
				pointer = i;
			} else {
				pointer = i + 1;
				return string("=");
			}
			break;
		case '\"':
			in_dq = !in_dq;
			break;
		case '\r':
		case '\n':
			break;
		default:
			ret += str[i];
			pointer = i;
			break;
		}
	}

	return ret;
}

// 񤭴Ūɬ
// äȼȴ顦
int UsogresRC::get_next_counter()
{
#ifndef WIN32
	const char* LOCKFILE = "seqno.lock";
	const char* COUNTERFILE = "counter";
	int tmpfd;
	char fn_buf[256];
	char lock_fn_buf[256];
	char counter_fn_buf[256];

	// I must rewrite the code below, check!! security!!
	sprintf(fn_buf, "%s/%s.%d", get_logdir().c_str(), LOCKFILE, getpid());
	sprintf(lock_fn_buf, "%s/%s", get_logdir().c_str(), LOCKFILE);
	sprintf(counter_fn_buf, "%s/%s", get_logdir().c_str(), COUNTERFILE);

	if ((tmpfd = creat(fn_buf, 0444)) < 0) {
		return -1;
	}
	close(tmpfd);

	while (link(fn_buf, lock_fn_buf) < 0) {
		if (errno != EEXIST) {
			return -1;
		}
		sleep(1);
	}

	if (unlink(fn_buf) < 0) {
		return -1;
	}

	FILE* fp;
	int cnt;
	if ((fp = fopen(counter_fn_buf, "r")) == NULL) {
		cnt = 0;
	} else {
		if (fscanf(fp, "%d", &cnt) == EOF) {
			cnt = 0;
		}
		fclose(fp);
	}
	if ((fp = fopen(counter_fn_buf, "w")) != NULL) {
		fprintf(fp, "%d\n", ++cnt);
		fclose(fp);
	}

	if (unlink(lock_fn_buf) < 0) {
		return -1;
	}

	return cnt;
#else
	return 0;
#endif
}
